//
// Author: Benjamin Langmann <b.langmann@gmx.de>, (C) 2005
//
// Copyright: See COPYING file that comes with this distribution
//
//

#include "stdafx.h"
#include "lmjpegVfW.h"
#include "lmjpeg_encoder.h"
#include "lmjpeg_io.h"
#include "resource.h"

HINSTANCE g_hInst;

void DPRINTF(int num, char *fmt, ...) {
	/*FILE *dfile;
	va_list args;
	char buf[10028];
	//if (num<4)return;
	//return ;

	va_start(args, fmt);
	vsprintf(buf, fmt, args);
	dfile=fopen("c:\\lmdebug.txt","a");
	fwrite(buf,1,strlen(buf),dfile);
	fwrite("\n",1,1,dfile);
	fclose(dfile);*/
}

BOOL WINAPI DllMain( HANDLE hModule, DWORD  ul_reason_for_call,LPVOID lpReserved) {
	g_hInst = (HINSTANCE) hModule;
    return TRUE;
}

LMJPEGVFW_API LRESULT WINAPI DriverProc(DWORD dwDriverId, HDRVR hDriver, UINT uMsg, LPARAM lParam1, LPARAM lParam2) {
	ICOPEN *icopen;
	ICINFO *icinfo;
	CODEC *codec = (CODEC*)dwDriverId;

	switch(uMsg)
	{
	/* driver primitives */
	case DRV_LOAD :
	case DRV_FREE :
		return DRVCNF_OK;

	case DRV_OPEN :
		DPRINTF(3,"DRV_OPEN");
		icopen = (ICOPEN *)lParam2;
		if (icopen != NULL && icopen->fccType != ICTYPE_VIDEO) {
			return DRVCNF_CANCEL;
		}
		
		codec=setup_codec();

		if (icopen != NULL) {
			icopen->dwError = ICERR_OK;
		}
		return (LRESULT)codec;

	case DRV_CLOSE :
		DPRINTF(3,"DRV_CLOSE");
		lmjpeg_end_coder(codec);
		end_codec(codec);
		return DRVCNF_OK;

	case DRV_DISABLE :
	case DRV_ENABLE :
		return DRVCNF_OK;

	case DRV_INSTALL :
	case DRV_REMOVE :
		return DRVCNF_OK;

	case DRV_QUERYCONFIGURE :
	case DRV_CONFIGURE :
		return DRVCNF_CANCEL;

	/* info */

	case ICM_GETINFO :
		DPRINTF(3,"ICM_GETINFO");
		icinfo = (ICINFO *)lParam1;
		icinfo->dwSize = (DWORD) lParam2;
		icinfo->fccType = ICTYPE_VIDEO;
		icinfo->fccHandler = FOURCC_LMJP;
		icinfo->dwFlags = VIDCF_COMPRESSFRAMES |
						  VIDCF_FASTTEMPORALC  |
						  VIDCF_FASTTEMPORALD ;
		icinfo->dwVersion = 0;
		icinfo->dwVersionICM = 0x0104;
		wcscpy(icinfo->szName, LMJPEG_NAME); 
		wcscpy(icinfo->szDescription, LMJPEG_DESC);
		return lParam2; /* size of struct */

		/* state control */

	case ICM_ABOUT :
		DPRINTF(3,"ICM_ABOUT");
		DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_ABOUT), (HWND)lParam1, about_proc, 0);
		return ICERR_OK;

	case ICM_CONFIGURE :
		DPRINTF(3,"ICM_CONFIGURE");
		if (lParam1 != -1)
		{	
			CODEC *co=setup_codec();
			memcpy(co->config, codec->config, sizeof(lmjpeg_fileheader));
			co->save=0;
			DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_CONFIG), (HWND)lParam1, configure_proc, (LPARAM)co);
			if (co->save) {
				memcpy(codec->config, co->config, sizeof(lmjpeg_fileheader));
			}
			end_codec(co);
		}
		return ICERR_OK;
			
	case ICM_GETSTATE :
		DPRINTF(3,"ICM_GETSTATE");
		if ((void*)lParam1 == NULL) {
			return sizeof(lmjpeg_fileheader);
		}
		memcpy((void*)lParam1, codec->config, sizeof(lmjpeg_fileheader));
		return ICERR_OK;

	case ICM_SETSTATE :
		DPRINTF(3,"ICM_SETSTATE");
		if ((void*)lParam1 == NULL) {
			DPRINTF(3,"ICM_SETSTATE : DEFAULT");
			end_codec(codec);
			codec=(CODEC*) setup_codec();
			return 0;
		}
		memcpy(codec->config,(void*)lParam1, sizeof(lmjpeg_fileheader));
		return 0; /* sizeof(CONFIG); */

	/* not sure the difference, private/public data? */

	case ICM_GET :
	case ICM_SET :
		return ICERR_OK;


	/* older-stype config */

	case ICM_GETDEFAULTQUALITY :
	case ICM_GETQUALITY :
	case ICM_SETQUALITY :
	case ICM_GETBUFFERSWANTED :
	case ICM_GETDEFAULTKEYFRAMERATE :
		return ICERR_UNSUPPORTED;

	/* compressor */

	case ICM_COMPRESS_QUERY :
		DPRINTF(3,"ICM_COMPRESS_QUERY");
		return compress_query(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_COMPRESS_GET_FORMAT :
		DPRINTF(3,"ICM_COMPRESS_GET_FORMAT");
		return compress_get_format(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_COMPRESS_GET_SIZE :
		DPRINTF(3,"ICM_COMPRESS_GET_SIZE");
		return compress_get_size(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_COMPRESS_FRAMES_INFO :
		DPRINTF(3,"ICM_COMPRESS_FRAMES_INFO");
		return compress_frames_info(codec, (ICCOMPRESSFRAMES *)lParam1);

	case ICM_COMPRESS_BEGIN :
		DPRINTF(3,"ICM_COMPRESS_BEGIN");
		return compress_begin(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_COMPRESS_END :
		DPRINTF(3,"ICM_COMPRESS_END");
		return compress_end(codec);

	case ICM_COMPRESS :
		DPRINTF(3,"ICM_COMPRESS");
		return compress_frame(codec, (ICCOMPRESS *)lParam1);

	/* decompressor */
	
	case ICM_DECOMPRESS_QUERY :
		DPRINTF(3,"ICM_DECOMPRESS_QUERY");
		return decompress_query(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_DECOMPRESS_GET_FORMAT :
		DPRINTF(3,"ICM_DECOMPRESS_GET_FORMAT");
		return decompress_get_format(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);
	
	case ICM_DECOMPRESS_BEGIN :
		DPRINTF(3,"ICM_DECOMPRESS_BEGIN");
		return decompress_begin(codec, (BITMAPINFO *)lParam1, (BITMAPINFO *)lParam2);

	case ICM_DECOMPRESS_END :
		DPRINTF(3,"ICM_DECOMPRESS_END");
		return decompress_end(codec);

	case ICM_DECOMPRESS :
		DPRINTF(3,"ICM_DECOMPRESS");
		return decompress_frame(codec, (ICDECOMPRESS *)lParam1);

	case ICM_DECOMPRESS_GET_PALETTE :
	case ICM_DECOMPRESS_SET_PALETTE :
	case ICM_DECOMPRESSEX_QUERY:
	case ICM_DECOMPRESSEX_BEGIN:
	case ICM_DECOMPRESSEX_END:
	case ICM_DECOMPRESSEX:
		return ICERR_UNSUPPORTED;

	default:
		if (uMsg < DRV_USER)
			return DefDriverProc(dwDriverId, hDriver, uMsg, lParam1, lParam2);
		else 
			return ICERR_UNSUPPORTED;
	}
}

void WINAPI Configure(HWND hwnd, HINSTANCE hinst, LPTSTR lpCmdLine, int nCmdShow) {
	DWORD dwDriverId = (DWORD) DriverProc(0, 0, DRV_OPEN, 0, 0);
	if (&dwDriverId != NULL) {
		if (lstrcmpi(lpCmdLine, "about")==0)
			DriverProc(dwDriverId, 0, ICM_ABOUT, (LPARAM)GetDesktopWindow(), 0);
		else
			DriverProc(dwDriverId, 0, ICM_CONFIGURE, (LPARAM)GetDesktopWindow(), 0);

		DriverProc(dwDriverId, 0, DRV_CLOSE, 0, 0);
	}
}

/* compressor */

LRESULT compress_query(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	BITMAPINFOHEADER * inhdr = &lpbiInput->bmiHeader;
	BITMAPINFOHEADER * outhdr = &lpbiOutput->bmiHeader;
	if (lpbiOutput == NULL)
		return ICERR_OK;

	if (inhdr->biWidth != outhdr->biWidth || inhdr->biHeight != outhdr->biHeight || outhdr->biCompression != FOURCC_LMJP)
		return ICERR_BADFORMAT;

	return ICERR_OK;
}

LRESULT compress_get_format(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	BITMAPINFOHEADER *inhdr = &lpbiInput->bmiHeader;
	BITMAPINFOHEADER *outhdr = &lpbiOutput->bmiHeader;
	if (lpbiOutput == NULL)
		return sizeof(BITMAPINFOHEADER);

	memcpy(outhdr, inhdr, sizeof(BITMAPINFOHEADER));
	outhdr->biSize = sizeof(BITMAPINFOHEADER);
	outhdr->biSizeImage = (DWORD) compress_get_size(codec, lpbiInput, lpbiOutput);
	outhdr->biXPelsPerMeter = 0;
	outhdr->biYPelsPerMeter = 0;
	outhdr->biClrUsed = 0;
	outhdr->biClrImportant = 0;
	outhdr->biCompression = FOURCC_LMJP;
	return ICERR_OK;
}

LRESULT compress_get_size(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	if (lpbiOutput->bmiHeader.biHeight>0)
		codec->config->height=lpbiOutput->bmiHeader.biHeight;
	else
		codec->config->height=-lpbiOutput->bmiHeader.biHeight;
	codec->config->width=lpbiOutput->bmiHeader.biWidth;

	// may have to request biggest possible pixel format (RGB)
	set_image_properties(codec);

	if (codec->config->jpeg_quality!=100)
		return 5*2*get_image_size(codec)/4;	// uint16_t
	else
		return 5*get_image_size(codec)/4;	// uint8_t
}

LRESULT compress_frames_info(CODEC *codec, ICCOMPRESSFRAMES *icf) {
	if (codec!=NULL) {
		codec->config->time_base_num = icf->dwScale;
		codec->config->time_base_den = icf->dwRate;
	}
	return ICERR_OK;
}

LRESULT compress_begin(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	if (codec!=NULL) {
		codec->config->width = lpbiInput->bmiHeader.biWidth;
		if (lpbiInput->bmiHeader.biHeight>0) {
			codec->config->height = lpbiInput->bmiHeader.biHeight;
			codec->config->topdown=0;
		}
		else {
			codec->config->height = -lpbiInput->bmiHeader.biHeight;
			codec->config->topdown=1;
		}
		codec->in_out_pix_format=LMJPEG_BGR;
		lmjpeg_setup_coder(codec);

		if (lpbiInput->bmiHeader.biCompression!=BI_RGB)
			return ICERR_BADFORMAT;
	}
	return ICERR_OK;
}


LRESULT compress_end(CODEC *codec) {
	lmjpeg_end_coder(codec);
	return ICERR_OK;
}

LRESULT compress_frame(CODEC *codec, ICCOMPRESS *icc) {
	lmjpeg_frameheader fheader;
	int size=0;
	uint8_t *tmp=(uint8_t*) icc->lpOutput+sizeof(lmjpeg_frameheader);
	fheader.frame_num=icc->lFrameNum;
	if (fheader.frame_num==0) {	// put fileheader before first frame
		memcpy(tmp, codec->config, sizeof(lmjpeg_fileheader));
		tmp+=sizeof(lmjpeg_fileheader);
		size+=sizeof(lmjpeg_fileheader);
	}
	
	// encode frame
	fheader.size=lmjpeg_encode(codec, (uint8_t*) icc->lpInput, tmp);
	
	fheader.marker=LMJPEG_IO_FRAMEHEADER | codec->frame_type;
	icc->lpbiOutput->biSizeImage=fheader.size+sizeof(lmjpeg_frameheader)+size;
	if (codec->frame_type==LMJPEG_IO_KEYFRAME) *icc->lpdwFlags = AVIIF_KEYFRAME;

	// put frame header befor frame data
	memcpy((uint8_t*) icc->lpOutput, &fheader, sizeof(lmjpeg_frameheader));
	codec->frame_num++; // or icc->lFrameNum;
	return ICERR_OK;
}

/* decompressor */


LRESULT decompress_query(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	BITMAPINFOHEADER * inhdr = &lpbiInput->bmiHeader;
	BITMAPINFOHEADER * outhdr = &lpbiOutput->bmiHeader;
	if (lpbiInput == NULL) {
		return ICERR_ERROR;
	}
	if (inhdr->biCompression != FOURCC_LMJP) {
		return ICERR_BADFORMAT;
	}
	if (lpbiOutput == NULL) {
		return ICERR_OK;
	}
	if (inhdr->biWidth != outhdr->biWidth || inhdr->biHeight != outhdr->biHeight \
		|| outhdr->biCompression!=BI_RGB \
		|| outhdr->biBitCount!=24) {
	//	|| (outhdr->biBitCount!=16 && outhdr->biBitCount!=24 && outhdr->biBitCount!=32)) {
		return ICERR_BADFORMAT;
	}
	return ICERR_OK;
}

// unused ???
LRESULT decompress_get_format(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	BITMAPINFOHEADER * inhdr = &lpbiInput->bmiHeader;
	BITMAPINFOHEADER * outhdr = &lpbiOutput->bmiHeader;
	LRESULT result;
	int height;
	if (lpbiOutput == NULL)
		return sizeof(BITMAPINFOHEADER);

	result = decompress_query(codec, lpbiInput, lpbiOutput);
	if (result != ICERR_OK)
		return result;
	
	if (inhdr->biHeight>0)
		height = inhdr->biHeight;
	else
		height = -inhdr->biHeight;
	outhdr->biSize = sizeof(BITMAPINFOHEADER);
	outhdr->biWidth = inhdr->biWidth;
	outhdr->biHeight = inhdr->biHeight;
	outhdr->biSizeImage = 3*height*inhdr->biWidth;
	outhdr->biPlanes = 1;
	outhdr->biBitCount = 24;
	outhdr->biCompression = BI_RGB;
	outhdr->biXPelsPerMeter = 0;
	outhdr->biYPelsPerMeter = 0;
	outhdr->biClrUsed = 0;
	outhdr->biClrImportant = 0;
	return ICERR_OK;
}

LRESULT decompress_begin(CODEC *codec, BITMAPINFO *lpbiInput, BITMAPINFO *lpbiOutput) {
	if (codec!=NULL) {
		BITMAPINFOHEADER * inhdr = &lpbiInput->bmiHeader;
		BITMAPINFOHEADER * outhdr = &lpbiOutput->bmiHeader;

		codec->config->width = inhdr->biWidth;
		if (inhdr->biHeight>0) {
			codec->config->height = inhdr->biHeight;
			codec->want_topdown=0;
		}
		else {
			codec->config->height = -inhdr->biHeight;
			codec->want_topdown=1;
		}
		codec->in_out_pix_format=LMJPEG_BGR;
		codec->frame_num=0;
	}
	return ICERR_OK;
}


LRESULT decompress_end(CODEC *codec) {
	lmjpeg_end_coder(codec);
	return ICERR_OK;
}

LRESULT decompress_frame(CODEC *codec, ICDECOMPRESS *icd) {
	lmjpeg_frameheader *fheader=icd->lpInput;
	uint8_t *tmp=(uint8_t*) icd->lpInput+sizeof(lmjpeg_frameheader);

	DPRINTF(5,"sizeimage: %d", icd->lpbiOutput->biSizeImage);
	DPRINTF(5,"bitcount: %d", icd->lpbiOutput->biBitCount);
	DPRINTF(5,"height: %d", icd->lpbiOutput->biHeight);
	DPRINTF(5,"width: %d", icd->lpbiOutput->biWidth);
	DPRINTF(5,"size: %d", icd->lpbiOutput->biSize);

	if (icd->lpbiInput->biCompression != FOURCC_LMJP \
		|| (fheader->marker & LMJPEG_IO_FRAMEHEADERMASK)!=LMJPEG_IO_FRAMEHEADER) {
		DPRINTF(5,"Error: decompress FOURCC not LMJP!");
		return ICERR_OK;
	}
	if (fheader->frame_num==0) {
		memcpy(codec->config, tmp, sizeof(lmjpeg_fileheader));
		if (codec->config->marker!=LMJPEG_IO_FILEHEADER) {
			DPRINTF(5,"Error: decompress wrong file header!");
			return ICERR_OK;
		}
		tmp+=sizeof(lmjpeg_fileheader);

		if (lmjpeg_setup_coder(codec) == -1)
			DPRINTF(5,"Error: decompress : lmjpeg_setup_coder!");
	}
	codec->frame_type=fheader->marker & ~LMJPEG_IO_FRAMEHEADERMASK;

	if (lmjpeg_decode(codec, (uint8_t*) icd->lpOutput, tmp, fheader->size) == -1)
		DPRINTF(5,"Error: decode!");
	codec->frame_num++;
	return ICERR_OK;
}

// -------------------------
// Dialog control functions
// -------------------------

BOOL CALLBACK about_proc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		case WM_INITDIALOG :
			break;
		case WM_COMMAND :
			if (LOWORD(wParam) == IDC_OK) {
				EndDialog(hDlg, LOWORD(wParam));
			}
			break;
		default :
			return 0;
	}
	return 1;
}

BOOL CALLBACK configure_proc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int sel;
	char str[20];
	CODEC *codec=(CODEC*) GetWindowLong(hDlg, GWL_USERDATA);
	switch (uMsg) {
		case WM_INITDIALOG :
			SetWindowLong(hDlg, GWL_USERDATA, lParam);
			codec=(CODEC*) lParam;
			SendDlgItemMessage(hDlg, IDC_MODE, CB_ADDSTRING, 0, (LPARAM)"Number of differing pixels");
			SendDlgItemMessage(hDlg, IDC_MODE, CB_ADDSTRING, 0, (LPARAM)"Sum of differences");
			if (codec->config->quality_mode==LMJPEG_MODE_NUMPIX) {
				SendDlgItemMessage(hDlg, IDC_MODE, CB_SETCURSEL, 0,0);
				EnableWindow(GetDlgItem(hDlg, IDC_NUM),TRUE);
			} else {
				SendDlgItemMessage(hDlg, IDC_MODE, CB_SETCURSEL, 1,0);
				EnableWindow(GetDlgItem(hDlg, IDC_NUM),FALSE);
			}
			sprintf(str,"%d",codec->config->quality_num);
			SetDlgItemText(hDlg, IDC_NUM, str);
			sprintf(str,"%d",codec->config->quality_diff);
			SetDlgItemText(hDlg, IDC_DIFF, str);

			sprintf(str,"%d", 70);
			SetDlgItemText(hDlg, IDC_JPEGQ, str);
			sprintf(str,"%d", 4);
			SetDlgItemText(hDlg, IDC_JPEGQ2, str);
			if (codec->config->jpeg_quality>=0 && codec->config->jpeg_quality<=100) {
				sprintf(str,"%d", codec->config->jpeg_quality);
				SetDlgItemText(hDlg, IDC_JPEGQ, str);
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_percent);
			}
			else if (codec->config->jpeg_quality>100 && codec->config->jpeg_quality<=131) {
				sprintf(str,"%.2f", codec->config->jpeg_quality-100.0);
				SetDlgItemText(hDlg, IDC_JPEGQ2, str);
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_factor);
			}
			else {
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_manual);
			}

			sprintf(str,"%d",codec->config->keyframe_intervall);
			SetDlgItemText(hDlg, IDC_KEYF, str);

			if (codec->config->dyn_keyframes)
				SendDlgItemMessage(hDlg, IDC_DYNAMIC, BM_SETCHECK, BST_CHECKED, 0); 
			else
				SendDlgItemMessage(hDlg, IDC_DYNAMIC, BM_SETCHECK, BST_UNCHECKED, 0);

			SendDlgItemMessage(hDlg, IDC_COMPRESS, CB_ADDSTRING, 0, (LPARAM)"RLE + Huffman Enc.");
			SendDlgItemMessage(hDlg, IDC_COMPRESS, CB_ADDSTRING, 0, (LPARAM)"Compress with zlib");
			if (codec->config->compression==LMJPEG_IO_RLE) {
				SendDlgItemMessage(hDlg, IDC_COMPRESS, CB_SETCURSEL, 0, 0);
			} else {
				SendDlgItemMessage(hDlg, IDC_COMPRESS, CB_SETCURSEL, 1, 0);
			}

			SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_ADDSTRING, 0, (LPARAM)"RGB");
			SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_ADDSTRING, 0, (LPARAM)"YUV2");
			SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_ADDSTRING, 0, (LPARAM)"YV12");
			if (codec->config->pix_format==LMJPEG_RGBP) {
				SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_SETCURSEL, 0, 0);
			} 
			else if (codec->config->pix_format==LMJPEG_YUV2) {
				SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_SETCURSEL, 1, 0);
			}
			else {
				SendDlgItemMessage(hDlg, IDC_COLORSPACE, CB_SETCURSEL, 2, 0);
			}

			break;
		case WM_COMMAND :
			if (LOWORD(wParam) == IDC_OK) {
				sel = (int) SendMessage(GetDlgItem(hDlg, IDC_MODE), CB_GETCURSEL, 0, 0);
				if (sel==0) {
					codec->config->quality_mode=LMJPEG_MODE_NUMPIX;
				}
				if (sel==1) {
					codec->config->quality_mode=LMJPEG_MODE_SUMDIFF;
				}
				codec->config->quality_num=GetDlgItemInt(hDlg, IDC_NUM, NULL, TRUE);
				codec->config->quality_diff=GetDlgItemInt(hDlg, IDC_DIFF, NULL, TRUE);
				codec->config->keyframe_intervall=GetDlgItemInt(hDlg, IDC_KEYF, NULL, TRUE);
				
				sel = (int) SendMessage(GetDlgItem(hDlg, IDC_COMPRESS), CB_GETCURSEL, 0, 0);
				if (sel==0) {
					codec->config->compression=LMJPEG_IO_RLE;
				}
				if (sel==1) {
					codec->config->compression=LMJPEG_IO_ZLIB;
				}
				
				sel = (int) SendMessage(GetDlgItem(hDlg, IDC_DYNAMIC), BM_GETCHECK, 0, 0);
				if (sel == BST_CHECKED) codec->config->dyn_keyframes=1;
				else codec->config->dyn_keyframes=0;

				sel = (int) SendMessage(GetDlgItem(hDlg, IDC_COLORSPACE), CB_GETCURSEL, 0, 0);
				if (sel==0) {
					codec->config->pix_format=LMJPEG_RGBP;
				}
				else if (sel==1) {
					codec->config->pix_format=LMJPEG_YUV2;
				}
				else if (sel==2) {
					codec->config->pix_format=LMJPEG_YV12;
				}

				sel = IsDlgButtonChecked(hDlg, IDC_jpeg_q_percent);
				if (sel==BST_CHECKED) {
					codec->config->jpeg_quality=GetDlgItemInt(hDlg, IDC_JPEGQ, NULL, TRUE);
					if (codec->config->jpeg_quality>=0 && codec->config->jpeg_quality<=100) {
						lmjpeg_calc_QM(codec);
					}
					else {
						MessageBox(hDlg, "Bad value! Should be >=0 and <=100.", "Warning", 0);
						sprintf(str,"%d",70);
						SetDlgItemText(hDlg, IDC_JPEGQ, str);
						codec->config->jpeg_quality=70;
						lmjpeg_calc_QM(codec);
					}
				}
				sel = IsDlgButtonChecked(hDlg, IDC_jpeg_q_factor);
				if (sel==BST_CHECKED) {
					GetDlgItemText(hDlg, IDC_JPEGQ2, str, 20);
					if (atoi(str)>0 && atoi(str)<=31) {
						codec->config->jpeg_quality=atoi(str)+100;
					}
					else {
						MessageBox(hDlg, "Bad value! Should be >=1 and <=31.", "Warning", 0);
						sprintf(str,"%d",4);
						SetDlgItemText(hDlg, IDC_JPEGQ2, str);
						codec->config->jpeg_quality=104;
					}
				}
				sel = IsDlgButtonChecked(hDlg, IDC_jpeg_q_manual);
				if (sel==BST_CHECKED) {
					codec->config->jpeg_quality=132;
				}

				if (check_config(codec->config)) codec->save=1;
				else {
					codec->save=0;
					MessageBox(hDlg, "Wrong values, will not save them!", "Warning", 0);
				}
				EndDialog(hDlg, LOWORD(wParam));
			}
			if (LOWORD(wParam) == IDC_CANCEL) {
				codec->save=0;
				EndDialog(hDlg, LOWORD(wParam));
			}
			if (LOWORD(wParam) == IDC_MODE) {
				sel = (int) SendMessage(GetDlgItem(hDlg, IDC_MODE), CB_GETCURSEL, 0, 0);
				if (sel==0) {
					EnableWindow(GetDlgItem(hDlg, IDC_NUM),TRUE);
				}
				if (sel==1) {
					EnableWindow(GetDlgItem(hDlg, IDC_NUM),FALSE);
				}
			}
			if (LOWORD(wParam) == IDC_MAT) {
				DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_MATRIXEDIT), hDlg, matrix_edit_proc, (LPARAM)codec->config->ratings);
			}
			if (LOWORD(wParam) == IDC_JPEGQ) {
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_percent);
			}
			if (LOWORD(wParam) == IDC_JPEGQ2) {
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_factor);	
			}
			if (LOWORD(wParam) == IDC_QM) {
				CheckRadioButton(hDlg, IDC_jpeg_q_percent, IDC_jpeg_q_manual, IDC_jpeg_q_manual);
				DialogBoxParam(g_hInst, MAKEINTRESOURCE(IDD_MATRIXEDIT), hDlg, matrix_edit_proc, (LPARAM)codec->config->myqm);
			}
			break;
		default :
			return 0;
	}
	return 1;
}

BOOL CALLBACK matrix_edit_proc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	int i, j, k;
	char str[1000],*tmp;
	uint8_t *mat=(uint8_t*) GetWindowLong(hDlg, GWL_USERDATA);
	switch (uMsg) {
		case WM_INITDIALOG :
			SetWindowLong(hDlg, GWL_USERDATA, lParam);
			mat=(uint8_t*) lParam;
			sprintf(str,"");
			for (i=0;i<8;i++) {
				sprintf(str,"%s%d",str,mat[i*8]);
				for (j=1;j<8;j++) {
					sprintf(str,"%s; %d",str,mat[i*8+j]);
				}
				sprintf(str,"%s\r\n",str);
			}
			SetDlgItemText(hDlg, IDC_MATRIX, str);
			break;
		case WM_COMMAND :
			if (LOWORD(wParam) == IDC_EDOK) {
				GetDlgItemText(hDlg, IDC_MATRIX, str, 1000);
				tmp=strtok(str,";\r\n");
				for (i=0;i<8;i++) {
					for (j=0;j<8;j++) {
						if (tmp!=NULL) {
							k=atoi(tmp);
							if (k<0) k=0;
							if (k>255) k=255;
							mat[i*8+j]=k;
						}
						tmp=strtok(NULL,";\r\n");
					}
				}
				EndDialog(hDlg, LOWORD(wParam));
			}
			if (LOWORD(wParam) == IDC_EDCANCEL) {
				EndDialog(hDlg, LOWORD(wParam));
			}
			break;
		default :
			return 0;
	}
	return 1;
}
