

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>

#include "../liblmjpeg/lmjpeg_encoder.h"
#include "../liblmjpeg/lmjpeg_io.h"
#include "../liblmjpeg/lmjpeg_image.h"
#include "../liblmjpeg/helpers.h"

#define OUTPUT_PIX_FORMAT PIX_FMT_YUV420P

int mywrite(CODEC *codec, char *in, char *out) {
	AVFormatContext *pFormatCtx;
	AVCodecContext  *pCodecCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame, *pFrame_conv;
	AVPacket        packet;
	int             j=0, i=0, videoStream, frameFinished, numBytes, av_pix_format, rle_size=0;
	uint8_t         *buffer=NULL;
	uint16_t	*rle_img;
	
	// pixel format mapping
	if (codec->in_out_pix_format==LMJPEG_YV12) av_pix_format=PIX_FMT_YUV420P;
	else if (codec->in_out_pix_format==LMJPEG_YUV2) av_pix_format=PIX_FMT_YUV422P;
	else if (codec->in_out_pix_format==LMJPEG_RGB) av_pix_format=PIX_FMT_RGB24;
	else if (codec->in_out_pix_format==LMJPEG_BGR) av_pix_format=PIX_FMT_BGR24;
	else {
		printf("Error: Wrong pixel format!\n");
		exit(1);
	}

	// Register all formats and codecs
	av_register_all();

	// Open video file
	if(av_open_input_file(&pFormatCtx, in, NULL, 0, NULL)!=0) return -1; // Couldn't open file

	// Retrieve stream information
	if(av_find_stream_info(pFormatCtx)<0) return -1; // Couldn't find stream information

	// Dump information about file onto standard error
	dump_format(pFormatCtx, 0, in, 0);

	// Find the first video stream
	videoStream=-1;
	for(i=0; i<pFormatCtx->nb_streams; i++)
		if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO) {
			videoStream=i;
			break;
		}
	if(videoStream==-1) return -1; // Didn't find a video stream

	// Get a pointer to the codec context for the video stream
	pCodecCtx=pFormatCtx->streams[videoStream]->codec;

	// Find the decoder for the video stream
	pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
	if(pCodec==NULL) return -1; // Codec not found

	// Open codec
	if(avcodec_open(pCodecCtx, pCodec)<0) return -1; // Could not open codec

	// Allocate video frame
	pFrame=avcodec_alloc_frame();
	
	// Allocate an AVFrame structure
	pFrame_conv=avcodec_alloc_frame();
	if(pFrame_conv==NULL) return -1;
	numBytes=avpicture_get_size(av_pix_format, pCodecCtx->width, pCodecCtx->height);
	buffer=(uint8_t*) malloc(sizeof(uint8_t)*numBytes);
	// Assign appropriate parts of buffer to image planes in pFrameRGB
	avpicture_fill((AVPicture *)pFrame_conv, buffer, av_pix_format, pCodecCtx->width, pCodecCtx->height);
	
	codec->config->width=pCodecCtx->width;
	codec->config->height=pCodecCtx->height;
	codec->config->time_base_num=pCodecCtx->time_base.num;
	codec->config->time_base_den=pCodecCtx->time_base.den;
	strcpy(codec->filename, out);
	lmjpeg_writer_init(codec);
	if (lmjpeg_setup_coder(codec)==-1) {
		printf("lmjpeg_setup_coder: something went wrong!\n");
		lmjpeg_writer_close(codec);
		lmjpeg_end_coder(codec); 
		free(buffer);
		av_free(pFrame_conv);
		av_free(pFrame);
		avcodec_close(pCodecCtx);
		av_close_input_file(pFormatCtx);
		exit(0);
	}
	rle_img=(uint16_t*) malloc(5*get_image_size(codec)*sizeof(uint16_t)/4);

	timing_data tdata;
	timing2k(&tdata, 0, "av: read frame");
	while(av_read_frame(pFormatCtx, &packet)>=0) { // && i<1) {
		uint8_t *ptr=packet.data;
		int len=packet.size;
		
		if(packet.stream_index==videoStream) { // Is this a packet from the video stream?
			while (len > 0) {
				// Decode video frame
				timing2k(&tdata, 1, "av: decode frame");
				numBytes=avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, ptr, len);
				j++;
				
				// Did we get a video frame?
				if(frameFinished) {
					timing2k(&tdata, 2, "av: convert frame");
					img_convert((AVPicture*) pFrame_conv, av_pix_format, (AVPicture*) pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);
				}
				else {
					//printf("-------- frame fragmented! ------------\n");
					//printf("Frame %d: got video packet %d: size: %d used: %d fnum: %d\n", i, j, len, numBytes,pCodecCtx->frame_number);
					// just do the last frame again
				}
				
				if (i>0 || frameFinished) {
					timing2k(&tdata, 3, "lmjpeg: encode frame");
					rle_size=lmjpeg_encode(codec, pFrame_conv->data[0], rle_img);
					timing2k(&tdata, 4, "lmjpeg: write frame");
					lmjpeg_write(codec, i, (uint8_t*) rle_img, rle_size); //pFrame->display_picture_number
					//if ((i%10)==0) printf("%d ",i);
					i++;
				}
				ptr += numBytes;
				len -= numBytes;
			}
		}
		// Free the packet that was allocated by av_read_frame
		av_free_packet(&packet);
		timing2k(&tdata, 0, "av: read frame");
	}
	
	if (1) {	// xvid may have a 1 frame delay
		// Decode video frame
		timing2k(&tdata, 1, "av: decode frame");
		numBytes=avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, NULL, 0);
				
		// Did we get a video frame?
		if(frameFinished) {
			timing2k(&tdata, 2, "av: convert frame");
			img_convert((AVPicture*) pFrame_conv, av_pix_format, (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);

			timing2k(&tdata, 3, "lmjpeg: encode frame");
			rle_size=lmjpeg_encode(codec, pFrame_conv->data[0], rle_img);
			timing2k(&tdata, 4, "lmjpeg: write frame");
			lmjpeg_write(codec, i, (uint8_t*) rle_img, rle_size);
			i++;
		}
	}
	
	printf("%d frames done!\n", i);
	lmjpeg_writer_close(codec);
	lmjpeg_end_coder(codec); 
	timing2k(&tdata, -1, NULL);
	
	free(rle_img);
	// Free the RGB image
	free(buffer);
	av_free(pFrame_conv);
	// Free the YUV frame
	av_free(pFrame);
	// Close the codec
	avcodec_close(pCodecCtx);
	// Close the video file
	av_close_input_file(pFormatCtx);
	return 0;
}


/* add a video output stream */
AVStream *add_video_stream(AVFormatContext *oc, int codec_id, int width, int height, int num, int den, int bitrate, int max_keyframe) {
	AVCodecContext *c;
	AVStream *st= av_new_stream(oc, 0);
	if (!st) {
		fprintf(stderr, "Could not alloc stream\n");
		exit(1);
	}
	c = st->codec;
	c->codec_id = CODEC_ID_MPEG4;//CODEC_ID_XVID; //codec_id;
	c->codec_type = CODEC_TYPE_VIDEO;
	c->bit_rate = bitrate;
	c->width = width;
	c->height = height;
	c->time_base.den = den;
	c->time_base.num = num;
	c->gop_size = max_keyframe; /* emit one intra frame every twelve frames at most */
	c->pix_fmt = OUTPUT_PIX_FORMAT;
	
	// some formats want stream headers to be seperate
	if(!strcmp(oc->oformat->name, "mp4") || !strcmp(oc->oformat->name, "mov") || !strcmp(oc->oformat->name, "3gp")) c->flags |= CODEC_FLAG_GLOBAL_HEADER;
	return st;
}

int myread(CODEC *codec, char *in, char *out, int bitrate, int max_keyframe) {
	uint8_t *rle_img;
	uint8_t *uncompressed_img=NULL;
	int i, size=-1, out_size, ret, av_pix_format;
	int video_outbuf_size = 200000;
	uint8_t *video_outbuf;
	
	AVOutputFormat *fmt;
	AVFormatContext *oc;
	AVStream *video_st=NULL; //*audio_st=NULL
	AVFrame *picture, *tmp_picture=NULL;

	// pixel format mapping
	if (codec->in_out_pix_format==LMJPEG_YV12) av_pix_format=PIX_FMT_YUV420P;
	else if (codec->in_out_pix_format==LMJPEG_YUV2) av_pix_format=PIX_FMT_YUV422P;
	else if (codec->in_out_pix_format==LMJPEG_RGB) av_pix_format=PIX_FMT_RGB24;
	else if (codec->in_out_pix_format==LMJPEG_BGR) av_pix_format=PIX_FMT_BGR24;
	//else if (codec->in_out_pix_format==LMJPEG_RGB32) av_pix_format=PIX_FMT_RGBA32;
	//else if (codec->in_out_pix_format==LMJPEG_RGB555) av_pix_format=PIX_FMT_RGB555;
	else {
		printf("Error: Wrong pixel format!\n");
		exit(1);
	}
	lmjpeg_frameheader frameheader;
	strcpy(codec->filename, in);
	lmjpeg_reader_init(codec);	// need width, height, ime_base_den / num
	
	/* initialize libavcodec, and register all codecs and formats */
	av_register_all();

	/* auto detect the output format from the name. default is mpeg. */
	fmt = guess_format(NULL, out, NULL);
	if (!fmt) {
		printf("Could not deduce output format from file extension: using MPEG.\n");
		fmt = guess_format("mpeg", NULL, NULL);
	}
	if (!fmt) {
		fprintf(stderr, "Could not find suitable output format\n");
		exit(1);
	}

	/* allocate the output media context */
	oc = av_alloc_format_context();
	if (!oc) {
		fprintf(stderr, "Memory error\n");
		exit(1);
	}
	oc->oformat = fmt;
	snprintf(oc->filename, sizeof(oc->filename), "%s", out);

	/* add the audio and video streams using the default format codecs and initialize the codecs */
	if (fmt->video_codec != CODEC_ID_NONE) {
		video_st = add_video_stream(oc, fmt->video_codec, codec->config->width, codec->config->height, codec->config->time_base_num, codec->config->time_base_den, bitrate, max_keyframe);
	}

	/* set the output parameters (must be done even if no parameters). */
	if (av_set_parameters(oc, NULL) < 0) {
		fprintf(stderr, "Invalid output format parameters\n");
		exit(1);
	}
	dump_format(oc, 0, out, 1);

	AVCodecContext *c=video_st->codec;
	/* now that all the parameters are set, we can open the audio and
	video codecs and allocate the necessary encode buffers */
	if (video_st) {
		AVCodec *avcodec;
		/* find the video encoder */
		avcodec = avcodec_find_encoder(c->codec_id);
		if (!codec) {
			fprintf(stderr, "codec not found\n");
			exit(1);
		}
		/* open the codec */
		if (avcodec_open(c, avcodec) < 0) {
			fprintf(stderr, "could not open codec\n");
			exit(1);
		}
		
		picture=avcodec_alloc_frame();
		size = avpicture_get_size(PIX_FMT_YUV420P, codec->config->width, codec->config->height);
		uint8_t *picture_buf= malloc(size);
		avpicture_fill((AVPicture *) picture, picture_buf, PIX_FMT_YUV420P, codec->config->width, codec->config->height);
			
		if (codec->in_out_pix_format==LMJPEG_YV12) {
			uncompressed_img=picture_buf;
		}
		else {
			tmp_picture=avcodec_alloc_frame();
			size = avpicture_get_size(av_pix_format, codec->config->width, codec->config->height);
			uncompressed_img=(uint8_t*) malloc(size);
			avpicture_fill((AVPicture *) tmp_picture, uncompressed_img, av_pix_format, codec->config->width, codec->config->height);
		}
		video_outbuf = (uint8_t*) malloc(video_outbuf_size);
	}
	
	/* open the output file, if needed */
	if (!(fmt->flags & AVFMT_NOFILE)) {
		if (url_fopen(&oc->pb, out, URL_WRONLY) < 0) {
			fprintf(stderr, "Could not open '%s'\n", out);
			exit(1);
		}
	}
			
	/* write the stream header, if any */
	av_write_header(oc);
	rle_img=(uint8_t*) malloc(3*5*codec->config->width*codec->config->height/4); // just in case ...

	timing_data tdata;
	timing2k(&tdata, 0,"lmjpeg: read frame");
	while ((size=lmjpeg_read(codec, &frameheader, rle_img))>=0) {
		timing2k(&tdata, 1, "lmjpeg: decode frame");
		lmjpeg_decode(codec, uncompressed_img, rle_img, size);
		
		timing2k(&tdata, 2, "av: convert frame");
		if (codec->in_out_pix_format!=LMJPEG_YV12) {
			img_convert((AVPicture*) picture, PIX_FMT_YUV420P, (AVPicture*) tmp_picture, av_pix_format, codec->config->width, codec->config->height);
		}
		
		timing2k(&tdata, 3, "av: encode frame");
		out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, picture);
		
		/* if zero size, it means the image was buffered */
		timing2k(&tdata, 4, "av: write frame");
		if (out_size > 0) {
			AVPacket pkt;
			av_init_packet(&pkt);

			pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, video_st->time_base);
			if(c->coded_frame->key_frame) pkt.flags |= PKT_FLAG_KEY;
			pkt.stream_index=video_st->index;
			pkt.data= video_outbuf;
			pkt.size= out_size;

			/* write the compressed frame in the media file */
			ret = av_write_frame(oc, &pkt);
		} 
		else ret = 0;
		if (ret != 0) {
			fprintf(stderr, "Error while writing video frame\n");
			exit(1);
		}
		timing2k(&tdata, 0, "lmjpeg: read frame");
	}
	timing2k(&tdata, -1, NULL);
	lmjpeg_end_coder(codec);
	printf("Done %d frames!\n",frameheader.frame_num+1);
	free(rle_img);
	
	/* close each codec */
	if (video_st) {
		avcodec_close(video_st->codec);
		av_free(picture->data[0]);
		av_free(picture);
		if (tmp_picture) {
			if (codec->in_out_pix_format!=LMJPEG_YV12) av_free(tmp_picture->data[0]);
			av_free(tmp_picture);
		}
		av_free(video_outbuf);
	}
	//if (audio_st) close_audio(oc, audio_st);

	/* write the trailer, if any */
	av_write_trailer(oc);

	/* free the streams */
	for(i = 0; i < oc->nb_streams; i++) {
		av_freep(&oc->streams[i]);
	}

	if (!(fmt->flags & AVFMT_NOFILE)) {
		/* close the output file */
		url_fclose(&oc->pb);
	}

	/* free the stream */
	av_free(oc);
	lmjpeg_reader_close(codec);	
	return 0;
}

int mywrite_xvid(CODEC *codec, char *in, char *out, int bitrate, int max_keyframe) {
	AVFormatContext *pFormatCtx;
	AVCodecContext  *pCodecCtx;
	AVCodec         *pCodec;
	AVFrame         *pFrame, *pFrame_conv;
	AVPacket        packet;
	int             j=0, i=0, videoStream, frameFinished, numBytes;
	uint8_t         *buffer=NULL;
	
	// for writing -----------
	int out_size, ret;
	int video_outbuf_size = 200000;
	uint8_t *video_outbuf;
	
	AVOutputFormat *fmt;
	AVFormatContext *oc;
	AVStream *video_st=NULL; //*audio_st=NULL	
	//-----------------------
	
	// Register all formats and codecs
	av_register_all();

	// Open video file
	if(av_open_input_file(&pFormatCtx, in, NULL, 0, NULL)!=0) return -1; // Couldn't open file

	// Retrieve stream information
	if(av_find_stream_info(pFormatCtx)<0) return -1; // Couldn't find stream information

	// Dump information about file onto standard error
	dump_format(pFormatCtx, 0, in, 0);

	// Find the first video stream
	videoStream=-1;
	for(i=0; i<pFormatCtx->nb_streams; i++) {
		if(pFormatCtx->streams[i]->codec->codec_type==CODEC_TYPE_VIDEO) {
			videoStream=i;
			break;
		}
	}
	if(videoStream==-1) return -1; // Didn't find a video stream

	// Get a pointer to the codec context for the video stream
	pCodecCtx=pFormatCtx->streams[videoStream]->codec;

	// Find the decoder for the video stream
	pCodec=avcodec_find_decoder(pCodecCtx->codec_id);
	if(pCodec==NULL) return -1; // Codec not found

	// Open codec
	if(avcodec_open(pCodecCtx, pCodec)<0) return -1; // Could not open codec

	// Allocate video frame
	pFrame=avcodec_alloc_frame();
	// Allocate an AVFrame structure
	pFrame_conv=avcodec_alloc_frame();
	if(pFrame_conv==NULL) return -1;
	
	numBytes=avpicture_get_size(PIX_FMT_YUV420P, pCodecCtx->width, pCodecCtx->height);
	buffer=(uint8_t*) malloc(sizeof(uint8_t)*numBytes);
	// Assign appropriate parts of buffer to image planes in pFrameRGB
	avpicture_fill((AVPicture *)pFrame_conv, buffer, PIX_FMT_YUV420P, pCodecCtx->width, pCodecCtx->height);
	
	// output ----------
	
	fmt = guess_format(NULL, out, NULL);
	if (!fmt) {
		printf("Could not deduce output format from file extension: using MPEG.\n");
		fmt = guess_format("mpeg", NULL, NULL);
	}
	if (!fmt) {
		fprintf(stderr, "Could not find suitable output format\n");
		exit(1);
	}

	/* allocate the output media context */
	oc = av_alloc_format_context();
	if (!oc) {
		fprintf(stderr, "Memory error\n");
		exit(1);
	}
	oc->oformat = fmt;
	snprintf(oc->filename, sizeof(oc->filename), "%s", out);

    /* add the audio and video streams using the default format codecs
	and initialize the codecs */
	if (fmt->video_codec != CODEC_ID_NONE) {
		video_st = add_video_stream(oc, fmt->video_codec, pCodecCtx->width, pCodecCtx->height, pCodecCtx->time_base.num, pCodecCtx->time_base.den, bitrate, max_keyframe);
	}

    /* set the output parameters (must be done even if no
	parameters). */
	if (av_set_parameters(oc, NULL) < 0) {
		fprintf(stderr, "Invalid output format parameters\n");
		exit(1);
	}
	dump_format(oc, 0, out, 1);

	AVCodecContext *c=video_st->codec;
    /* now that all the parameters are set, we can open the audio and
	video codecs and allocate the necessary encode buffers */
	if (video_st) {
		AVCodec *avcodec;
		/* find the video encoder */
		avcodec = avcodec_find_encoder(c->codec_id);
		if (!codec) {
			fprintf(stderr, "codec not found\n");
			exit(1);
		}
		/* open the codec */
		if (avcodec_open(c, avcodec) < 0) {
			fprintf(stderr, "could not open codec\n");
			exit(1);
		}
		video_outbuf = (uint8_t*) malloc(video_outbuf_size);
	}
	
	/* open the output file, if needed */
	if (!(fmt->flags & AVFMT_NOFILE)) {
		if (url_fopen(&oc->pb, out, URL_WRONLY) < 0) {
			fprintf(stderr, "Could not open '%s'\n", out);
			exit(1);
		}
	}
			
	/* write the stream header, if any */
	av_write_header(oc);
	
	// ----------
	
	timing_data tdata;
	timing2k(&tdata, 0, "av: read frame");
	while(av_read_frame(pFormatCtx, &packet)>=0) {
		uint8_t *ptr=packet.data;
		int len=packet.size;
		
		if(packet.stream_index==videoStream) { // Is this a packet from the video stream?
			while (len > 0) {
				// Decode video frame
				timing2k(&tdata, 1, "av: decode frame");
				numBytes=avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, ptr, len);
				
				// Did we get a video frame?
				if(frameFinished) {
					timing2k(&tdata, 2, "av: convert frame");
					img_convert((AVPicture*) pFrame_conv, PIX_FMT_YUV420P, (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);
				}
				else {
					printf("-------- frame fragmented! ------------\n");
					printf("Frame %d: got video packet %d: size: %d used: %d\n", i, j, packet.size, numBytes);
				}
				
				if (frameFinished) {
					timing2k(&tdata, 3, "av: encode frame");
					out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, pFrame_conv);
					
					timing2k(&tdata, 4, "av: write frame");
					if (out_size > 0) {
						AVPacket pkt;
						av_init_packet(&pkt);
	
						pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, video_st->time_base);
						if(c->coded_frame->key_frame) pkt.flags |= PKT_FLAG_KEY;
						pkt.stream_index=video_st->index;
						pkt.data= video_outbuf;
						pkt.size= out_size;
	
						/* write the compressed frame in the media file */
						ret = av_write_frame(oc, &pkt);
					} 
					else ret = 0;
					if (ret != 0) {
						fprintf(stderr, "Error while writing video frame\n");
						exit(1);
					}
					i++;
				}
				ptr += numBytes;
				len -= numBytes;
			}
			j++;
		}
		//else {
		//	printf("read other stream: %d\n", packet.stream_index);	
		//}
		
		// Free the packet that was allocated by av_read_frame
		av_free_packet(&packet);
		timing2k(&tdata, 0, "av: read frame");
	}
	
	if(1) {	// xvid may have a 1 frame delay
		// Decode video frame
		timing2k(&tdata, 1, "av: decode frame");
		numBytes=avcodec_decode_video(pCodecCtx, pFrame, &frameFinished, NULL, 0);
		
		// Did we get a video frame?
		if(frameFinished) {
			timing2k(&tdata, 2, "av: convert frame");
			img_convert((AVPicture*) pFrame_conv, PIX_FMT_YUV420P, (AVPicture*)pFrame, pCodecCtx->pix_fmt, pCodecCtx->width, pCodecCtx->height);
					
			timing2k(&tdata, 3, "av: encode frame");
			out_size = avcodec_encode_video(c, video_outbuf, video_outbuf_size, pFrame_conv);
					
			timing2k(&tdata, 4, "av: write frame");
			if (out_size > 0) {
				AVPacket pkt;
				av_init_packet(&pkt);
	
				pkt.pts= av_rescale_q(c->coded_frame->pts, c->time_base, video_st->time_base);
				if(c->coded_frame->key_frame) pkt.flags |= PKT_FLAG_KEY;
				pkt.stream_index=video_st->index;
				pkt.data= video_outbuf;
				pkt.size= out_size;
	
				/* write the compressed frame in the media file */
				ret = av_write_frame(oc, &pkt);
			} 
			else ret = 0;
			if (ret != 0) {
				fprintf(stderr, "Error while writing video frame\n");
				exit(1);
			}
			i++;
		}
	}
	timing2k(&tdata, -1, NULL);
	printf("%d frames done!\n", i);

	// Free the RGB image
	free(buffer);
	av_free(pFrame_conv);
	// Free the YUV frame
	av_free(pFrame);
	// Close the codec
	avcodec_close(pCodecCtx);
	// Close the video file
	av_close_input_file(pFormatCtx);
	
	/* close each codec */
	if (video_st) {
		avcodec_close(video_st->codec);
		av_free(video_outbuf);
	}
	
	/* write the trailer, if any */
	av_write_trailer(oc);

	/* free the streams */
	for(i = 0; i < oc->nb_streams; i++)
		av_freep(&oc->streams[i]);

	if (!(fmt->flags & AVFMT_NOFILE))
		url_fclose(&oc->pb); /* close the output file */

	/* free the stream */
	av_free(oc);
	return 0;
}

void usage() {
	printf("Usage: lmjpeg [options] read  in.lmjpeg out.avi\n");
	printf("              [options] write in.avi    out.lmjpeg\n");
	printf("Options for write:\n");
	printf("    -quality_mode x  : PIX : if more than y pixel of a block differ from\n");
	printf("                             the old block more than z use new block.\n");
	printf("                     : DIFF: if the pixel of a block differ together from\n");
	printf("                             the old block more than z use new block.\n");
	printf("    -quality y z     : y=0-64, z=0-255 default: y=z=5\n");
	printf("    -jpeg_quality x  : 1-100\n");
	printf("    -mpeg_quant x    : quantizer, 1 <= x <= 31\n");
	printf("    -keyframe x      : 0 (no keyframes) - ...\n");
	printf("    -dyn_keyframes   : set keyframes dynamically\n");
	printf("    -pix_format x    : internal pixel format: YV12, YUV2, RGB\n");
	printf("    -in_pix_format x : input pixel format: YV12, YUV2, RGB, BGR\n");
	printf("    -compression x   : ZLIB, RLE\n");
	printf("Options for read:\n");
	printf("    -bitrate          : eg. 1500\n");
	printf("    -keyframe x       : 0 (no keyframes) - ...\n");
	printf("    -out_pix_format x : pixel format for output: YV12, YUV2, RGB, BGR\n");
	printf("Options for encoding to xvid:\n");
	printf("    lmjpeg [options] xvid write in.avi out.avi\n");
	printf("    -bitrate          : eg. 1500\n");
	printf("    -keyframe x       : 0 (no keyframes) - ...\n");
	exit(0);
}

int main(int argc, char *argv[]) {
	int i;
	char *in=NULL,*out=NULL;
	int bitrate=700000;
	int max_keyframe=100;
	int action=0, tmp;
	int xvid=0;
	int got;

	timing_data tdata;
	timing2k(&tdata, 0,"coding time");
	
	CODEC *codec=setup_codec();
	for (i=1;i<argc;i++) {
		got=0;
		if (strncmp(argv[i],"-quality_mode",13)==0) {
			got=1;
			i++;
			if (i<argc) {
				if (strncmp(argv[i],"PIX",3)==0) codec->config->quality_mode = LMJPEG_MODE_NUMPIX;
				else if (strncmp(argv[i],"DIFF",4)==0) codec->config->quality_mode = LMJPEG_MODE_SUMDIFF;
				else usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-quality",8)==0) {
			got=1;
			i++;
			if (i+1<argc) {
				codec->config->quality_num=atoi(argv[i]);
				if (codec->config->quality_num<0) usage();
				if (codec->config->quality_num>64) usage();
				i++;
				codec->config->quality_diff=atoi(argv[i]);
				if (codec->config->quality_diff<0) usage();
				//if (codec->config->quality_diff>255) usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-mpeg_quant",11)==0) {
			got=1;
			i++;
			if (i<argc) {
				tmp=atoi(argv[i]);
				if (tmp<1) usage();
				if (tmp>31) usage();
				codec->config->jpeg_quality=tmp+100;
				lmjpeg_calc_QM(codec);
			}
			else usage();
		}
		if (strncmp(argv[i],"-jpeg_quality",13)==0) {
			got=1;
			i++;
			if (i<argc) {
				codec->config->jpeg_quality=atoi(argv[i]);
				if (codec->config->jpeg_quality<0) usage();
				if (codec->config->jpeg_quality>100) usage();
				lmjpeg_calc_QM(codec);
			}
			else usage();
		}
		
		if (strncmp(argv[i],"-keyframe",9)==0) {
			got=1;
			i++;
			if (i<argc) {
				codec->config->keyframe_intervall=atoi(argv[i]);
				max_keyframe=codec->config->keyframe_intervall;
				if (codec->config->keyframe_intervall<0) usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-dyn_keyframes",14)==0) {
			got=1;
			codec->config->dyn_keyframes=1;
		}
		if (strncmp(argv[i],"-pix_format",11)==0) {
			got=1;
			i++;
			if (i<argc) {
				if (strncmp(argv[i],"YV12",4)==0) codec->config->pix_format= LMJPEG_YV12;
				else if (strncmp(argv[i],"YUV2",4)==0) codec->config->pix_format= LMJPEG_YUV2;
					else if (strncmp(argv[i],"RGB",5)==0) codec->config->pix_format= LMJPEG_RGBP;
						else usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-in_pix_format",11)==0) {
			got=1;
			i++;
			if (i<argc) {
				if (strncmp(argv[i],"YV12",4)==0) codec->in_out_pix_format= LMJPEG_YV12;
				else if (strncmp(argv[i],"YUV2",4)==0) codec->in_out_pix_format= LMJPEG_YUV2;
					//else if (strncmp(argv[i],"RGB32",5)==0) codec->in_out_pix_format= LMJPEG_RGB32;
					//	else if (strncmp(argv[i],"RGB555",6)==0) codec->in_out_pix_format= LMJPEG_RGB555;
							else if (strncmp(argv[i],"RGB",3)==0) codec->in_out_pix_format= LMJPEG_RGB;
								else if (strncmp(argv[i],"BGR",3)==0) codec->in_out_pix_format= LMJPEG_BGR;
									else usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-out_pix_format",11)==0) {
			got=1;
			i++;
			if (i<argc) {
				if (strncmp(argv[i],"YV12",4)==0) codec->in_out_pix_format= LMJPEG_YV12;
				else if (strncmp(argv[i],"YUV2",4)==0) codec->in_out_pix_format= LMJPEG_YUV2;
					//else if (strncmp(argv[i],"RGB32",5)==0) codec->in_out_pix_format= LMJPEG_RGB32;
					//	else if (strncmp(argv[i],"RGB555",6)==0) codec->in_out_pix_format= LMJPEG_RGB555;
							else if (strncmp(argv[i],"RGB",3)==0) codec->in_out_pix_format= LMJPEG_RGB;
								else if (strncmp(argv[i],"BGR",3)==0) codec->in_out_pix_format= LMJPEG_BGR;
									else usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-bitrate",8)==0) {
			got=1;
			i++;
			if (i<argc) {
				bitrate=atoi(argv[i])*1024;
				if (bitrate<1025) usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"-compression",12)==0) {
			got=1;
			i++;
			if (i<argc) {
				
				if (strncmp(argv[i],"RLE",3)==0) codec->config->compression = LMJPEG_IO_RLE;
				else if (strncmp(argv[i],"ZLIB",4)==0) codec->config->compression = LMJPEG_IO_ZLIB;
				else usage();
			}
			else usage();
		}
		if (strncmp(argv[i],"read",4)==0) {
			got=1;
			if (i+2<argc) {
				action=1;
				in=argv[++i];
				out=argv[++i];
			}
			else usage();
		}
		if (strncmp(argv[i],"write",5)==0) {
			got=1;
			if (i+2<argc) {
				action=2;
				in=argv[++i];
				out=argv[++i];
			}
			else usage();
		}
		if (strncmp(argv[i],"xvid",4)==0) {
			got=1;
			xvid=1;
		}

		if (got==0) {
			printf("\nError: unrecognized option: %s\n\n", argv[i]);
			usage();
		}
	}
	
	if (action==0) usage();
	if (action==1) myread(codec, in, out, bitrate, max_keyframe);
	if (action==2) {
		if (xvid!=1) mywrite(codec, in, out);
		else mywrite_xvid(codec, in, out, bitrate, max_keyframe);
	}
	end_codec(codec);
	timing2k(&tdata, -1, NULL);
	return 0;
}
